//
//  AppDelegate.h
//  gemmlowp_test
//
//  Created by petewarden on 9/28/15.
//  Copyright (c) 2015 petewarden. All rights reserved.
//

#import <UIKit/UIKit.h>

@interface AppDelegate : UIResponder<UIApplicationDelegate>

@property(strong, nonatomic) UIWindow *window;

@end
